/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


#define EXTERN extern 
#define EXTERN_INIT(x,y) extern x

char* ASLType[4] = {"TypeCaractere","TypeEntier","TypeReel","TypeAdresse"} ;

#include "yoocom.hxx"
#include "yoocom_y.h"

SvcClass::SvcClass(SvcHead* h, ListPtr<SvcMember>* m_l) 
{
  this->head = h ;

  IterListPtr<SvcMember> iterM(*m_l) ;  
  for (; iterM.more(); iterM.next()) {
    if (iterM.get()->isLocal()) {
      this->attribut_list.insert((SvcLocalMember *)iterM.get(), 0) ;
    } else {
      this->method_list.insert((SvcRemoteMember *)iterM.get(), 0) ;
    }
  }
}


SvcStruct::SvcStruct(char* n, ListPtr<SvcMember>* m_l) 
{
  this->name = n ;
  if (m_l != NULL) { 
    attribut_list.capture(*m_l) ; 
  }
}


SvcRemoteMember::SvcRemoteMember(SvcType* r, char* n, ListPtr<SvcParam>* p_l)
{
  name = strdup(n) ;
  result = r ;
  if (p_l != NULL) { 
    parameter_list.capture(*p_l) ; 
  }
} 
   

int SvcRemoteMember::isLocal() 
{ 
  return 0 ; 
}


SvcLocalMember::SvcLocalMember(SvcType* t, char* n) : type(t)
{
  name = strdup(n) ;
}


SvcMultipleLocalMember::SvcMultipleLocalMember(SvcType* t, char* n, int s)
: SvcLocalMember(t, n), size(s) 
{
}


int SvcLocalMember::isLocal() 
{ 
  return 1 ; 
}


SvcParam::SvcParam(int io, SvcType* t, char* n) 
: direction(io), typ(t)
{
  if (n != NULL) {
    name = strdup(n) ;
  } else {
    name = NULL ;
  }
}


int SvcParam::in() 
{ 
  return direction == IN ; 
}


int SvcParam::out() 
{
  return direction != IN ; 
}


void SvcRemoteMember::cppGenRegister(int num, char* cl)
{
  fprintf(yyout, "\n  class ReqRegister_%s {\n", this->name) ;
  fprintf(yyout, "    public :\n") ;
  fprintf(yyout, "    int dummy ;\n") ;
  fprintf(yyout, "    ReqRegister_%s", this->name) ;
  fprintf(yyout, "(SvcServer* svc = SvcServer::_svcBase) {\n") ;
  fprintf(yyout, "      svc->reqRegister(%s::svc_%s,%d);\n",cl,this->name,num);
  fprintf(yyout, "    } \n") ;
  fprintf(yyout, "  } req_%s ;\n", this->name) ;
}


void SvcRemoteMember::cppGenServer(int num, char* cl)
{
  fprintf(yyout, "  static void svc_%s(SvcSocket& __svc) {\n", this->name) ;

  // Parameter declaration
  IterListPtr<SvcParam> iter(this->parameter_list) ;
  for (; iter.more(); iter.next()) {
    iter.get()->genType("    ") ;
    iter.get()->genName(" ") ;
    fprintf(yyout, " ;\n") ;
  }

  // IN parameter receive
  int param = 0 ;
  iter.reset() ;
  for (; iter.more(); iter.next()) {
    if (iter.get()->in()) {
      if (param++ == 0) fprintf(yyout, "    __svc") ;
      iter.get()->genName(" >> ") ;
    }
  }
  if (param != 0) fprintf(yyout, " ;\n") ;
  
  // Function call
  param = 0 ;
  if (this->result->tid() != T_VOID) {
    fprintf(yyout, "    __svc << ") ;
    param = 1 ;
  } else {
    fprintf(yyout, "    ") ;
  }
  fprintf(yyout, "( ((%s*)(&__svc))->%s(", cl, this->name) ;
  iter.reset() ;
  if (iter.more()) iter.get()->genName("") ;
  iter.next() ;
  for (; iter.more(); iter.next()) {
    iter.get()->genName(", ") ;
  }
  fprintf(yyout, ") )") ;

  iter.reset() ;  
  for (; iter.more(); iter.next()) {
    if (iter.get()->out()) {
      if (param++ == 0) fprintf(yyout, ";\n    __svc") ;
      iter.get()->genName(" << ") ;
    }
  }
  if (param != 0) {
    fprintf(yyout, " << svcflush ;\n") ;
  } else {
    fprintf(yyout, ";\n") ;
  }

  iter.reset() ;  
  for (; iter.more(); iter.next()) {
    if ( (iter.get()->in()) && (iter.get()->tid() == T_STRING) &&
	 !(iter.get()->istemplate()) && !(iter.get()->isflow()) &&
	 !(iter.get()->istable()) ) {
      iter.get()->genName("    delete ") ;
      fprintf(yyout, " ;\n") ;
    }
  }

  fprintf(yyout, "  } ;\n") ;

  this->cppGenRegister(num, cl) ;
}


void SvcRemoteMember::cppGenClient(int numproc, char* n, int mode) 
{
  this->result->genType("  ") ;
  fprintf(yyout, " %s(", name) ;
  IterListPtr<SvcParam> iter(this->parameter_list) ;
  if (iter.more()) {
    if ( (iter.get()->in()) && (iter.get()->issimple() == 0)) {
      fprintf(yyout, "const ") ;
    }
    iter.get()->genType("") ;
    if ( (iter.get()->out()) || (iter.get()->issimple() == 0)) {
      iter.get()->genName("& ") ;
    } else {
      iter.get()->genName(" ") ;
    }
    iter.next() ;
  }
  for (; iter.more(); iter.next()) {
    if ( (iter.get()->in()) && (iter.get()->issimple() == 0)) {
      fprintf(yyout, ", const ") ;
    } else {
      fprintf(yyout, ", ") ;
    }
    iter.get()->genType("") ;
    if ( (iter.get()->out()) || (iter.get()->issimple() == 0)) {
      iter.get()->genName("& ") ;
    } else {
      iter.get()->genName(" ") ;
    }
  }
  fprintf(yyout, ")\n  {\n    ") ;

  fprintf(yyout, "int numproc = %d ;\n    ", numproc) ;
  if (result->tid() != T_VOID) {
    result->genType("") ;
    fprintf(yyout, " result ;\n    ") ;
  }

  if (mode == DISCONNECTED) {
    fprintf(yyout, "this->reimport() ;\n    ") ;
  }

  fprintf(yyout, "(*(SvcSocket*)this) << numproc") ;
  iter.reset() ;
  for (; iter.more(); iter.next()) {
    if (iter.get()->in()) {
      iter.get()->genName(" << ") ;
    }
  }
  fprintf(yyout, " << svcflush ;\n") ;

  int param = 0 ;
  if (result->tid() != T_VOID) {
    fprintf(yyout, "    (*(SvcSocket*)this) >> result") ;
    param = 1 ; 
  } 
  iter.reset() ;
  for (; iter.more(); iter.next()) {
    if (iter.get()->out()) {
      if (param++ == 0) {
	fprintf(yyout, "    (*(SvcSocket*)this)") ;
      }
      iter.get()->genName(" >> ") ;
    }
  }

  if (param != 0) {
    fprintf(yyout, ";\n") ;
  }

  if (mode == DISCONNECTED) {
    fprintf(yyout, "    this->close() ;\n") ;
  }

  if (result->tid() != T_VOID) {
    fprintf(yyout, "    return result ;\n") ;
  }
  fprintf(yyout, "  } ;\n") ;
}


void SvcRemoteMember::aslGenServer(int num, char* cl)
{
  fprintf(yyout, "\nCopier (%s_methodes : %d)\n", cl, num) ;
  fprintf(yyout, "  Fonction [(this) var val]\n") ;
  if (result->isflow()) {
    fprintf(yyout, "  [\n    _SvcWriteStream this 0 %s_%s this", cl, this->name) ;
  } else if (result->istable()) {
    fprintf(yyout, "  [\n    _SvcWriteTable this %s_%s this", cl, this->name);
  } else if (result->tid() != T_VOID) {
    fprintf(yyout, "  [\n    _SvcWrite this %s_%s this", cl, this->name);
  } else {
    fprintf(yyout, "  [\n    %s_%s this", cl, this->name);
  }

  // Parameter list
  int flowcount = 1 ;
  IterListPtr<SvcParam> iter(this->parameter_list) ;
  for (; iter.more(); iter.next()) {
    if (iter.get()->isflow()) {
      fprintf(yyout, "\n      _SvcReadStream this %d ", flowcount++) ;
      iter.get()->genAslType("") ;
    } else if (iter.get()->istable()) {
      iter.get()->genAslType("\n      _SvcReadTable this ") ;
    } else {
      iter.get()->genAslType("\n      _SvcRead this ") ;
    }
  }

  if (result->tid() != T_VOID) {
    fprintf(yyout, "\n    _SvcSend this") ;
  }
  fprintf(yyout, "\n  ]") ;
}


void SvcRemoteMember::aslGenClient(int num, char* cl)
{
  if (result->isflow()) {
    fprintf(yyout, "Fonction [ mode_remanent ") ;
  } else {
    fprintf(yyout, "Fonction [ ") ;
  }
  fprintf(yyout, "%s_%s(this", cl, this->name) ;

  // Parameter list
  IterListPtr<SvcParam> iter(this->parameter_list) ;
  for (; iter.more(); iter.next()) {
    iter.get()->genName(",") ;
  }
  fprintf(yyout, ") var val]\n[\n") ;
  
  if (conn_gen == DISCONNECTED) {
    fprintf(yyout, "  _SvcReimport this\n") ;
  }

  char* tab ;
  if (this->result->isflow()) {
    fprintf(yyout, "\n  Quand (Statut == 1) [\n  ") ;
    tab = "  " ;
  } else {
    tab = "" ;
  }
  
  fprintf(yyout, "  _SvcWrite this %d\n", num) ;

  int flowcount = 1 ;
  iter.reset() ;
  for (; iter.more(); iter.next()) {
    fprintf(yyout, "%s", tab) ;
    if (iter.get()->istable()) {      
      iter.get()->genName("  _SvcWriteTable this ") ;
    } else if (iter.get()->isflow()) {
      iter.get()->genName("  TantQue (Non FinDeFlot? ") ;
      fprintf(yyout, ") [\n      _SvcWriteNext this %d ", flowcount) ;
      iter.get()->genName("") ;
      iter.get()->genName("\n      SuivantDe [") ;
      fprintf(yyout, "]\n    ]\n    _SvcWriteFlush this %d", flowcount++) ;
    } else {
      iter.get()->genName("  _SvcWrite this ") ;
    }
    fprintf(yyout, "\n") ;
  }

  fprintf(yyout, "%s  _SvcSend this",tab) ;

  if (this->result->isflow()) {
    this->result->genAslType("\n    _SvcReadFlush this 0 ") ;
    fprintf(yyout, "\n  ]") ;
    this->result->genAslType("\n  Donner [val] _SvcReadNext this 0 ") ;
  }
  
  if (conn_gen == DISCONNECTED) {
    fprintf(yyout, "\n  SvcClose this") ;
  }

  if (result->tid() != T_VOID) {
    if (result->isflow()) {
      fprintf(yyout, "\n  Quand (FinDeFlot? val) [ FinDeFlot ]") ;
      fprintf(yyout, "\n  Resultat val") ;
    } else if (result->istable()) {
      result->genAslType("\n  Resultat _SvcReadTable this ") ;
    } else {
      result->genAslType("\n  Resultat _SvcRead this ") ;
    }
  }

  fprintf(yyout, "\n]\n\n") ;
}


void SvcStruct::aslGen()
{
  fprintf(yyout, "Donner [%s] ",this->name) ;
  fprintf(yyout, "DeclarerStructure (\n") ;
  IterListPtr<SvcLocalMember> iter(this->attribut_list) ;
  if (iter.more()) {
    iter.get()->genName("  \"") ;
    iter.get()->genAslType("\", ") ;
    iter.next() ;
  }
  for (; iter.more(); iter.next()) {
    iter.get()->genName(",\n  \"") ;
    iter.get()->genAslType("\", ") ;
  }
  fprintf(yyout, "\n)\n\n\n") ;
}

void SvcClass::cppGenClient()
{
  int beginProc = this->head->base == NULL ? 0 : this->head->base->numproc() ;
  char* b = this->head->base==NULL?"SvcClient":this->head->base->classname() ;

  if (this->head->base == NULL) {
    fprintf(yyout, "class proxy_%s : public %s {\n", this->head->name, b) ;
    this->head->cppGenCtor(b, "proxy_") ;
  } else {
    fprintf(yyout, "class proxy_%s : public proxy_%s {\n", this->head->name,b);
    this->head->cppGenCtor(b, "proxy_") ;
  }


  if (conn_gen == DISCONNECTED) {
    fprintf(yyout, "  int import(char* filter) {\n") ;
    fprintf(yyout, "    if (this->SvcClient::import(filter)==-1) {\n") ;
    fprintf(yyout, "      return -1 ;\n    } else {\n");
    fprintf(yyout, "      return this->SvcClient::close() ;\n    }\n  }\n") ;
  }
  
  for (int i = 0; i < this->method_list.count(); i++) {
    method_list[i]->cppGenClient(i+beginProc, head->name, conn_gen) ;
  }
  fprintf(yyout, "} ; \n\n") ;
}


void SvcClass::cppGenServerRemote()
{
  int beginProc = this->head->base == NULL ? 0 : this->head->base->numproc() ;

  fprintf(yyout, "\n  public :\n") ;
  
  for (int i = 0; i < this->method_list.count(); i++) {
    this->method_list[i]->cppGenServer(i+beginProc, this->head->name) ;
  }
}


void SvcClass::aslGenServer()
{
  fprintf(yyout, "Donner [%s] ",this->head->name) ;
  fprintf(yyout, "DeclarerStructure (\n  [svcobj], TypeEntier") ;
  IterListPtr<SvcLocalMember> iter(this->attribut_list) ;
  for (; iter.more(); iter.next()) {
    iter.get()->genName(",\n  [") ;
    iter.get()->genAslType("], ") ;
  }
  fprintf(yyout, "\n)\n\n") ;

  fprintf(yyout, "Donner [%s_methodes] ", this->head->name) ;
  fprintf(yyout, "AllouerTable TypeFonction %d\n", this->method_list.count()) ;
  for (int i = 0; i < this->method_list.count(); i++) {
    this->method_list[i]->aslGenServer(i, this->head->name) ;
  }
  fprintf(yyout, "\n\n\n") ;
}


void SvcClass::aslGenClient()
{
  fprintf(yyout, "Donner [proxy_%s] ",this->head->name) ;
  fprintf(yyout, "DeclarerStructure (\n  [svcobj], TypeEntier") ;
  fprintf(yyout, "\n)\n\n") ;

  for (int i = 0; i < this->method_list.count(); i++) {
    this->method_list[i]->aslGenClient(i, this->head->name) ;
  }
  fprintf(yyout, "\n\n\n") ;
}


void SvcHead::cppGenCtor(char* b, char* pref) 
{
  fprintf(yyout, "\n  public :\n  ") ;
  fprintf(yyout, "%s%s(int family=AF_INET) :",pref, this->name) ;
  fprintf(yyout, " %s%s(\"%s\", family) {};\n  ", pref, b, this->name) ;
  fprintf(yyout, "%s%s(char* n, int f) : ", pref, this->name);
  fprintf(yyout, "%s%s(n,f) {}\n", pref, b) ;
}


void SvcType::genAslType(char* s)
{
  if (this->id != T_OTHER) {
    fprintf(yyout, "%s%s", s, ASLType[this->id]) ;
  } else {
    fprintf(yyout, "%s%s", s, this->typename) ;
  }
}


void cppGenServerHeader() 
{
  fprintf(yyout, " public SvcServer ") ;
}


SvcClass* findClass(char* n) 
{
  IterListPtr<SvcClass> iterC(SchemaClass) ;  
  for (; iterC.more(); iterC.next()) {
    if (strcmp(iterC.get()->head->name, n) == 0) {
      return iterC.get() ;
    }
  }
  return NULL ;
}


void cppGenClient() 
{
  IterListPtr<SvcClass> iterC(SchemaClass) ;  
  for (; iterC.more(); iterC.next()) {
    iterC.get()->cppGenClient() ;
  }
}


void aslGenStruct()
{
  IterListPtr<SvcStruct> iter(SchemaStruct) ;  
  for (; iter.more(); iter.next()) {
    iter.get()->aslGen() ;
  }
}

void aslGenClient()
{
  aslGenStruct() ;
  IterListPtr<SvcClass> iterC(SchemaClass) ;  
  for (; iterC.more(); iterC.next()) {
    iterC.get()->aslGenClient() ;
  }
}


void aslGenServer()
{
  aslGenStruct() ;
  IterListPtr<SvcClass> iterC(SchemaClass) ;  
  for (; iterC.more(); iterC.next()) {
    iterC.get()->aslGenServer() ;
  }
}
